# ------- #
# Imports #
# ------- #

from traceback import format_exc
from types import SimpleNamespace as o
from .utils import twoLineSeps
import importlib
import os


# ---- #
# Main #
# ---- #


def validateAndGetReportFn(reporter, silent, cliResult):
    validationResult = o(report=None, cliResult=cliResult, hasError=False)

    try:
        reporterModule = importlib.import_module(reporter)
    except:
        if not silent:
            err = "An error occurred while importing the reporter"
            cliResult.stderr = os.linesep + err + twoLineSeps + format_exc()

        cliResult.code = 2
        validationResult.hasError = True
        return validationResult

    if hasattr(reporterModule, "report") and callable(reporterModule.report):
        validationResult.report = reporterModule.report
    else:
        if not silent:
            cliResult.stderr = (
                os.linesep + "the reporter must expose a callable 'report'"
            )

        cliResult.code = 2
        validationResult.hasError = True

    return validationResult



from .internal.joinWith import joinWith


def toWrittenList(aList):
    andSeparated = aList[-2:]
    commaSeparated = aList[:-2]
    commaSeparated.append(joinWith(" and ")(andSeparated))
    return joinWith(", ")(commaSeparated)



from .utils import makeTestResults
from .. import chalk, green


def runTests():
    chainedResults = makeTestResults("chained styles")
    e = chainedResults.errors

    expected = f"\x1b[32;47;1;3msuccess\x1b[0m"
    actual = green.bgWhite.bold.underline("success")
    if actual != expected:
        e.append('green.bgWhite.bold.underline("success")')

    actual = chalk.green.bgWhite.bold.underline("success")
    if actual != expected:
        e.append('chalk.green.bgWhite.bold.underline("success")')

    expected = f"\x1b[31;44;1msuccess\x1b[0m"
    actual = green.bgWhite.bold.bgBlue.red("success")
    if actual != expected:
        e.append('green.bgWhite.bold.bgBlue.red("success")')

    actual = chalk.green.bgWhite.bold.bgBlue.red("success")
    if actual != expected:
        e.append('chalk.green.bgWhite.bold.bgBlue.red("success")')

    return [chainedResults]



from . import chainedStyles, colors, expectedErrors, miscStyles
from .utils import printResults
from ..src import utils as _


def runTests():
    testModules = [colors, miscStyles, chainedStyles, expectedErrors]

    _.forEach(_runTestModule)(testModules)


def _runTestModule(aTestModule):
    listOfResults = aTestModule.runTests()
    _.forEach(printResults)(listOfResults)



# ------- #
# Imports #
# ------- #

from . import minimal
from ..fns import forEach, isLaden
from .explainErrors import explainErrors
from .initRootState import initRootState


# ---- #
# Main #
# ---- #


def report(state, *, showErrorDetails=True):
    state = initRootState(state)

    rootTests = state.tests
    rootSuites = state.suites

    forEach(minimal.reportTest)(rootTests)

    if isLaden(rootTests) and isLaden(rootSuites):
        print()

    forEach(minimal.reportSuite)(rootSuites)

    if showErrorDetails and not state.succeeded:
        print(explainErrors(state))



from .all_ import all_
from .any_ import any_
from .appendOne import appendOne
from .applyOneTo import applyOneTo
from .assign import assign
from .discardFirst import discardFirst
from .discardWhen import discardWhen
from .endsWith import endsWith
from .forEach import forEach
from .getAttributeKeys import getAttributeKeys
from .getListOfCollectionKeys import getListOfCollectionKeys
from .getListOfCollectionValues import getListOfCollectionValues
from .iif import iif
from .invoke import invoke
from .invokeAttr import invokeAttr
from .isEmpty import isEmpty
from .isLaden import isLaden
from .joinWith import joinWith
from .keepWhen import keepWhen
from .map_ import map_
from .noop import noop
from .passThrough import passThrough
from .prependStr import prependStr
from .raise_ import raise_
from .split import split
from .toWrittenList import toWrittenList

__all__ = [
    "all_",
    "any_",
    "appendOne",
    "applyOneTo",
    "assign",
    "discardFirst",
    "discardWhen",
    "endsWith",
    "forEach",
    "getAttributeKeys",
    "getListOfCollectionKeys",
    "getListOfCollectionValues",
    "iif",
    "invoke",
    "invokeAttr",
    "isEmpty",
    "isLaden",
    "joinWith",
    "keepWhen",
    "map_",
    "noop",
    "passThrough",
    "prependStr",
    "raise_",
    "split",
    "toWrittenList",
]



# ------- #
# Imports #
# ------- #

from .. import *  # noqa: F403
from .utils import makeTestResults
from ..src import utils as _


# ---- #
# Init #
# ---- #

miscStyles = [("bold", 1), ("dim", 2), ("underline", 3), ("hidden", 6)]


# ---- #
# Main #
# ---- #


def runTests():
    miscStyleResults = makeTestResults("misc styles")

    miscStyleResults.errors = _.reduce(toResults, [])(miscStyles)

    return [miscStyleResults]


# ------- #
# Helpers #
# ------- #


def toResults(errors, style):
    name, code = style
    expected = f"\x1b[{code}msuccess\x1b[0m"
    if globals()[name]("success") != expected:
        errors.append(f'{name}("success")')
    if getattr(chalk, name)("success") != expected:  # noqa: F405
        errors.append(f'chalk.{name}("success")')

    return errors



from .assign import assign
from .discardWhen import discardWhen
from .forEach import forEach
from .iif import iif
from .isEmpty import isEmpty
from .isLaden import isLaden
from .joinWith import joinWith
from .mAppendAll import mAppendAll
from .mAppendAllTo import mAppendAllTo
from .mAssign import mAssign
from .mSet import mSet
from .map_ import map_
from .noop import noop
from .passThrough import passThrough
from .reduce import reduce
from .set_ import set_

__all__ = [
    "assign",
    "discardWhen",
    "forEach",
    "iif",
    "isEmpty",
    "isLaden",
    "joinWith",
    "mAppendAll",
    "mAppendAllTo",
    "mAssign",
    "mSet",
    "map_",
    "noop",
    "passThrough",
    "reduce",
    "set_",
]



# ------- #
# Imports #
# ------- #

from .internal.makeCallFn import makeCallFn
from .internal.getTypedResult import getTypedResult


# ---- #
# Init #
# ---- #

moduleName = "map_"


# ---- #
# Main #
# ---- #


def map_(mapperFn):
    callMapperFn = makeCallFn(mapperFn, 3, moduleName)

    def map_inner(collection):
        typedMap = getTypedResult(collection, _typeToMap, moduleName)
        return typedMap(callMapperFn, collection)

    return map_inner


# ------- #
# Helpers #
# ------- #


def map_list(callMapperFn, aList):
    result = []
    for idx, el in enumerate(aList):
        result.append(callMapperFn(el, idx, aList))

    return result


_typeToMap = {list: map_list}



# ------- #
# Imports #
# ------- #

from collections import OrderedDict

from .utils import isEmpty, throw


# ---- #
# Main #
# ---- #

#
# I'm wrapping the Chalk class in a factory method for cases where people want
#   their own instance clear of global polution (good for testing)
#
def makeAChalkClass():
    class Chalk:
        def __init__(self):
            setattr(self, "_styleCodes", OrderedDict())

        def _addStyle(self, styleType, code):
            self._styleCodes[styleType] = code
            return self

        def _getCodeList(self):
            return ";".join(self._styleCodes.values())

        def __call__(self, str):
            codeList = self._getCodeList()
            if isEmpty(codeList):
                throw(
                    ValueError,
                    f"""\
                You attempted to call the chalk instance without any styles.
                This is either due to a bug in 'simple_chalk' or you are
                expecting this library to work in a way I didn't intend.  Either
                way I'd appreciate if you raised an issue on github so I can
                prevent others from reaching this.
                    """,
                )

            return f"\x1b[{codeList}m{str}\x1b[0m"

        def _clone(self):
            result = type(self)()
            result._styleCodes.update(self._styleCodes)
            return result

    return Chalk



# ------- #
# Imports #
# ------- #

import re
from .fns import all_, any_, forEach, isEmpty as areEmpty, keepWhen, map_


# ---- #
# Main #
# ---- #

#
# This method mutates state
#


def onlyKeepGreppedTests(state, grepArgs):
    if all_(areEmpty)(grepArgs):
        return state

    grepArgs = map_(allToRegexes)(grepArgs)

    markGreppedTests(state, grepArgs.grepTests)
    markGreppedSuites(state, grepArgs.grepSuites)
    markGreppedTestsAndSuites(state, grepArgs.grep)
    trimNonMarkedTestsAndSuites(state)
    removeAllMarks(state)

    return state


# ------- #
# Helpers #
# ------- #


def mark(testOrSuite):
    testOrSuite._keep = True

    if testOrSuite.parentSuite:
        mark(testOrSuite.parentSuite)


def allToRegexes(grepStrings):
    return map_(lambda aString: re.compile(aString))(grepStrings)


def regexMatches(someString):
    def regexMatches_inner(aRegex):
        return bool(aRegex.search(someString))

    return regexMatches_inner


def markTestIfLabelMatches(regexes):
    def markTestIfLabelMatches_inner(testOrSuite):
        if any_(regexMatches(testOrSuite.label))(regexes):
            mark(testOrSuite)

    return markTestIfLabelMatches_inner


def markGreppedTests(stateOrSuite, grepTests):
    forEach(markTestIfLabelMatches(grepTests))(stateOrSuite.tests)
    forEach(lambda suite: markGreppedTests(suite, grepTests))(
        stateOrSuite.suites
    )


def markAllTestsIfLabelMatches(regexes):
    def markAllTestsIfLabelMatches_inner(suite):
        if any_(regexMatches(suite.label))(regexes):
            forEach(mark)(suite.tests)

    return markAllTestsIfLabelMatches_inner


def markGreppedSuites(stateOrSuite, grepSuites):
    forEach(markAllTestsIfLabelMatches(grepSuites))(stateOrSuite.suites)
    forEach(lambda suite: markGreppedSuites(suite, grepSuites))(
        stateOrSuite.suites
    )


def markGreppedTestsAndSuites(stateOrSuite, grep):
    forEach(markTestIfLabelMatches(grep))(stateOrSuite.tests)
    forEach(markAllTestsIfLabelMatches(grep))(stateOrSuite.suites)
    forEach(lambda suite: markGreppedTestsAndSuites(suite, grep))(
        stateOrSuite.suites
    )


def isMarked(testOrSuite):
    return getattr(testOrSuite, "_keep", False)


def trimNonMarkedTestsAndSuites(stateOrSuite):
    stateOrSuite.tests = keepWhen(isMarked)(stateOrSuite.tests)
    stateOrSuite.suites = keepWhen(isMarked)(stateOrSuite.suites)
    forEach(trimNonMarkedTestsAndSuites)(stateOrSuite.suites)


def removeMark(testOrSuite):
    delattr(testOrSuite, "_keep")


def removeAllMarks(stateOrSuite):
    forEach(removeMark)(stateOrSuite.tests)
    forEach(removeMark)(stateOrSuite.suites)
    forEach(removeAllMarks)(stateOrSuite.suites)



# ------- #
# Imports #
# ------- #

from .data import first8Colors, second8Colors
from .utils import invoke, passThrough
from .makeAChalkClass import makeAChalkClass
from .makeChalkHelpers import (
    makeStyles,
    makeOrderedStyles,
    makePrefixedOrderedStyles,
)


# ---- #
# Main #
# ---- #


def makeAChalkInstance():
    return passThrough(
        makeAChalkClass(),
        [
            invoke,
            _makeForegroundStyles,
            _makeBackgroundStyles,
            _addAliases,
            _makeNonColorStyles,
        ],
    )


# ------- #
# Helpers #
# ------- #


def _makeForegroundStyles(chalkInst):
    return makeOrderedStyles(
        chalkInst=chalkInst,
        styleListStartCodePairs=[(first8Colors, 30), (second8Colors, 90)],
        styleType="fg",
    )


def _makeBackgroundStyles(chalkInst):
    return makePrefixedOrderedStyles(
        chalkInst=chalkInst,
        prefix="bg",
        styleListStartCodePairs=[(first8Colors, 40), (second8Colors, 100)],
        styleType="bg",
    )


def _addAliases(chalkInst):
    return makeStyles(
        chalkInst,
        [
            ("gray", 90, "fg"),
            ("grey", 90, "fg"),
            ("bgGray", 100, "bg"),
            ("bgGrey", 100, "bg"),
        ],
    )


def _makeNonColorStyles(chalkInst):
    return makeStyles(
        chalkInst,
        [
            ("bold", 1, "bold"),
            ("dim", 2, "dim"),
            ("underline", 3, "underline"),
            #
            # leaving out "blink" (4) and "invert" (5) until a compelling
            #   use-case arises
            #
            ("hidden", 6, "hidden"),
        ],
    )



# ------- #
# Imports #
# ------- #

from .internal.makeCallFn import makeCallFn
from .internal.getTypedResult import getTypedResult


# ---- #
# Init #
# ---- #

moduleName = "mMap"


# ---- #
# Main #
# ---- #


def mMap(mapperFn):
    callMapperFn = makeCallFn(mapperFn, 3, moduleName)

    def mMap_inner(collection):
        typedMMap = getTypedResult(collection, _typeToMMap, moduleName)
        return typedMMap(callMapperFn, collection)

    return mMap_inner


# ------- #
# Helpers #
# ------- #


def mMap_list(callMapperFn, aList):
    for idx, el in enumerate(aList):
        aList[idx] = callMapperFn(el, idx, aList)

    return aList


_typeToMMap = {list: mMap_list}



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Init #
# ---- #

moduleName = "append"


# ---- #
# Main #
# ---- #


def append(el):
    def append_inner(collection):
        typedAppend = getTypedResult(collection, _typeToAppend, moduleName)
        return typedAppend(el, collection)

    return append_inner


# ------- #
# Helpers #
# ------- #


def append_list(el, aList):
    result = aList.copy().append(el)
    return result


_typeToAppend = {list: append_list}



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Init #
# ---- #

moduleName = "appendAll"


# ---- #
# Main #
# ---- #


def appendAll(collectionToAppend):
    def appendAll_inner(collection):
        typedAppendAll = getTypedResult(
            collection, _typeToAppendAll, moduleName
        )
        return typedAppendAll(collectionToAppend, collection)

    return appendAll_inner


# ------- #
# Helpers #
# ------- #


def appendAll_list(listToAppend, aList):
    return aList + listToAppend


_typeToAppendAll = {list: appendAll_list}



from types import SimpleNamespace as o

#
# README
#  - I have no idea what to name this.  Python just doesn't have a clean ternary
#    operator which stinks and I didn't want to use 'iif' because that's
#    unreadable.  So for now we have a verbose and less performant, but readable
#    function chain
#


def whenTruthy(condition):
    def return_(truthyResult):
        def otherwise(falseyResult):
            if condition:
                return truthyResult
            else:
                return falseyResult

        return o(otherwise=otherwise)

    return o(return_=return_)



# ------- #
# Imports #
# ------- #

from inspect import Parameter
from math import inf

from .getFnSignature import getFnSignature


# ---- #
# Init #
# ---- #

_nonVarPositionalParamKinds = {
    Parameter.POSITIONAL_ONLY,
    Parameter.POSITIONAL_OR_KEYWORD,
}


# ---- #
# Main #
# ---- #


def getNumPositionalParams(fn, callerName):
    sig = getFnSignature(fn, callerName)

    if _hasVarPositionalParam(sig):
        return inf

    numPositionalParams = 0
    for p in sig.parameters.values():
        if p.kind in _nonVarPositionalParamKinds:
            numPositionalParams += 1

    return numPositionalParams


# ------- #
# Helpers #
# ------- #


def _hasVarPositionalParam(sig):
    for p in sig.parameters.values():
        if p.kind is Parameter.VAR_POSITIONAL:
            return True



first8Colors = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
]

second8Colors = [
    "blackBright",
    "redBright",
    "greenBright",
    "yellowBright",
    "blueBright",
    "magentaBright",
    "cyanBright",
    "whiteBright",
]



from .append import append
from .appendAll import appendAll
from .assignMissing import assignMissing
from .flatten import flatten
from .forEach import forEach
from .get import get
from .getAttributeKeys import getAttributeKeys
from .invoke import invoke
from .invokeAttr import invokeAttr
from .isEmpty import isEmpty
from .isLaden import isLaden
from .isSomething import isSomething
from .isTruthy import isTruthy
from .join import join
from .keepWhen import keepWhen
from .lowerFirst import lowerFirst
from .mMap import mMap
from .map_ import map_
from .not_ import not_
from .passThrough import passThrough
from .reduce import reduce
from .set_ import set_
from .startsWith import startsWith
from .tee import tee
from .throw import throw
from .toString import toString
from .upperFirst import upperFirst
from .whenTruthy import whenTruthy

__all__ = [
    "append",
    "appendAll",
    "assignMissing",
    "flatten",
    "forEach",
    "get",
    "getAttributeKeys",
    "invoke",
    "invokeAttr",
    "isEmpty",
    "isLaden",
    "isSomething",
    "isTruthy",
    "join",
    "keepWhen",
    "lowerFirst",
    "mMap",
    "map_",
    "not_",
    "passThrough",
    "reduce",
    "set_",
    "startsWith",
    "tee",
    "throw",
    "toString",
    "upperFirst",
    "whenTruthy",
]



# ------- #
# Imports #
# ------- #

from ._vendor.po import case_conversion
from ._vendor.po.case_conversion import camelcase, dashcase
from copy import deepcopy
from types import SimpleNamespace as o

from .fns import (
    getListOfCollectionKeys,
    map_,
    passThrough,
    prependStr,
    raise_,
    toWrittenList,
)


# ---- #
# Init #
# ---- #

#
# case_conversion.dashcase takes more than a single argument so `map_` will pass
#   it the index and original array which may screw things up
#
def dashcase(aString):
    return case_conversion.dashcase(aString)


_grepArgs = o(grep=[], grepSuites=[], grepTests=[])
_grepArgsKeys = passThrough(
    _grepArgs, [getListOfCollectionKeys, map_(dashcase), map_(prependStr("--"))]
)
_availableGrepArgsKeys = toWrittenList(_grepArgsKeys)


# ---- #
# Main #
# ---- #


#
# We can assume the argument order
#   0    reporter
#   1    silent
#   2?+  grep | grepSuites | grepTests
#
#
# and we can also assume
#  - reporter is a non-relative module name
#  - silent is a string boolean
#  - grepArgs may or may not exist.  Validation is only for debugging purposes
#    as calling code should be reliable.
#
#
def parseArgs(*args):
    return o(reporter=args[0], silent=args[1], grepArgs=parseGrepArgs(args[2:]))


# ------- #
# Helpers #
# ------- #


def parseGrepArgs(grepArgs):
    result = deepcopy(_grepArgs)

    i = 0
    grepArgsLen = len(grepArgs)
    while i < grepArgsLen:
        arg = grepArgs[i]
        if arg not in _grepArgsKeys:
            raise_(
                ValueError,
                f"""
                key '{arg}' is invalid
                available grep keys: {_availableGrepArgsKeys}
                """,
            )

        grepKey = arg
        i += 1

        if i == grepArgsLen:
            raise_(
                ValueError,
                f"""
                a value must be given to '{grepKey}'
                """,
            )

        grepVals = getattr(result, camelcase(grepKey))
        grepVals.append(grepArgs[i])
        i += 1

    return result



# ------- #
# Imports #
# ------- #

from .internal.makeGenericCallFn import makeGenericCallFn
from .internal.getTypedResult import getTypedResult
from .decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def any_(predicate):
    fnName = any.__name__
    callPredicate = makeGenericCallFn(predicate, 3, fnName)

    def any_inner(collection):
        typedAny = getTypedResult(collection, typeToAny, fnName)
        return typedAny(callPredicate, collection)

    return any_inner


# ------- #
# Helpers #
# ------- #


def any_list(callPredicate, aList):
    for idx, el in enumerate(aList):
        if callPredicate(el, idx, aList):
            return True

    return False


typeToAny = {list: any_list}



from .decorators.argIsType import argIsType


@argIsType(str)
def endsWith(prefix):
    @argIsType(str)
    def endsWith_inner(fullStr):
        return fullStr.endswith(prefix)

    return endsWith_inner



class Style:
    def __init__(self, name, code, styleType):
        self.name = name
        self.code = code
        self.styleType = styleType

    def __get__(self, instance, _owner):
        result = instance._clone()
        result._addStyle(self.styleType, self.code)
        return result

    def __set__(self, instance, value):
        raise AttributeError(f"the attribute '{self.name}' cannot be set")



# ------- #
# Imports #
# ------- #

from .Style import Style
from .utils import flatten, map_, passThrough, reduce, upperFirst


# ---- #
# Main #
# ---- #


def makeStyles(chalkInst, styleInputList):
    def toStyles(result, styleIn):
        name, code, styleType = styleIn
        setattr(type(chalkInst), name, Style(name, str(code), styleType))
        return chalkInst

    return reduce(toStyles, chalkInst)(styleInputList)


def makeOrderedStyles(*, chalkInst, styleListStartCodePairs, styleType):
    return passThrough(
        styleListStartCodePairs,
        [map_(_toStyle(styleType)), flatten, reduce(_toChalk, chalkInst)],
    )


def makePrefixedOrderedStyles(
    *, chalkInst, prefix, styleListStartCodePairs, styleType
):
    prefixedPairs = map_(_addPrefix(prefix))(styleListStartCodePairs)
    return makeOrderedStyles(
        chalkInst=chalkInst,
        styleListStartCodePairs=prefixedPairs,
        styleType=styleType,
    )


# ------- #
# Helpers #
# ------- #


def _toChalk(chalkInst, aStyle):
    setattr(type(chalkInst), aStyle.name, aStyle)
    return chalkInst


def _toStyle(styleType):
    def _toStyle_inner(pair):
        styleList, startCode = pair

        def _makeAStyle(name, idx):
            return Style(name, str(idx + startCode), styleType)

        return map_(_makeAStyle)(styleList)

    return _toStyle_inner


def _addPrefix(prefix):
    def addPrefix_inner(pair):
        styleList, startCode = pair
        styleList = map_(lambda s: prefix + upperFirst(s))(styleList)
        return (styleList, startCode)

    return addPrefix_inner



# ------- #
# Imports #
# ------- #

from copy import copy
from ..fns import discardWhen, joinWith, map_, mAppendAllTo, passThrough, reduce
import os


# ---- #
# Init #
# ---- #

border = "-" * 20
twoLineSeps = os.linesep + os.linesep
separator = twoLineSeps + border + twoLineSeps


# ---- #
# Main #
# ---- #


def explainErrors(state):
    rootTestsCopy = passThrough(state.tests, [discardWhen(didSucceed), copy])

    flatTests = passThrough(
        state.suites,
        [discardWhen(didSucceed), reduce(flattenTests, rootTestsCopy)],
    )

    return passThrough(
        flatTests, [map_(toExplanation), joinWith(separator), wrapInBorder]
    )


# ------- #
# Helpers #
# ------- #


def toExplanation(aTest):
    explanation = []

    if aTest.parentSuite:
        explanation.append("suite: " + aTest.parentSuite.label)

    explanation.append("test: " + aTest.label)
    explanation.append("")
    explanation.append(str(aTest.error).strip())
    return os.linesep.join(explanation)


def flattenTests(result, aSuite):
    if hasattr(aSuite, "suites"):
        result = reduce(flattenTests, result)(aSuite.suites)

    return passThrough(
        aSuite.tests, [discardWhen(didSucceed), mAppendAllTo(result)]
    )


def didSucceed(testOrSuite):
    return testOrSuite.succeeded


def wrapInBorder(content):
    return (
        twoLineSeps
        + border
        + twoLineSeps
        + content
        + twoLineSeps
        + border
        + os.linesep
    )



#
# README
#  - This is like 'update' except it doesn't overwrite existing keys.  It only
#    assigns new ones.
#


def assignMissing(fromDict):
    def assignMissing_inner(intoDict):
        for k, v in fromDict.items():
            if k not in intoDict:
                intoDict[k] = v

        return intoDict

    return assignMissing_inner



from .decorators.argIsInstance import argIsInstance


@argIsInstance(str)
def prependStr(prependThis):
    fnName = prependStr.__name__

    @argIsInstance(str, fnName)
    def prependStr_inner(toThis):
        return prependThis + toThis

    return prependStr_inner



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace as o
from ..src import utils as _


# ---- #
# Init #
# ---- #

_x = "\x1b[31m✘\x1b[0m"
_o = "\x1b[32m✔\x1b[0m"


# ---- #
# Main #
# ---- #


def makeTestResults(label):
    return o(label=label, errors=[])


def printResults(testResults):
    errors = testResults.errors
    label = testResults.label

    if _.isLaden(errors):
        print(label)
        _.forEach(_printAResult)(errors)
    else:
        print(f"{_o} {label}")

    print()


# ------- #
# Helpers #
# ------- #


def _printAResult(result):
    print(f"  {_x} {result}")



# ------- #
# Imports #
# ------- #

from glob import glob
from os import path
from subprocess import run
from .suite import suite
from .test import test
import importlib.util
import os
import sys

from .fns import (
    discardWhen,
    endsWith,
    forEach,
    joinWith,
    map_,
    passThrough,
    raise_,
    split,
)


# ---- #
# Main #
# ---- #

twoLineSeps = os.linesep + os.linesep


#
# I'm choosing to gather all the tests prior to running any because I feel that
#   will be a simpler design.
#
def gatherTests(aSuite):
    oldCurrentSuite = aSuite.rootState.currentSuite
    aSuite.rootState.currentSuite = aSuite
    aSuite.fn()
    forEach(gatherTests)(aSuite.suites)
    aSuite.rootState.currentSuite = oldCurrentSuite


#
# I can't find a clean way to do this so I'm rolling my own.  The python
#   import system is inherently hacky anyway :(
#
def importTests():
    globStr = path.join("tests", "**", "*.py")

    passThrough(
        globStr,
        [
            recursiveGlob,
            discardWhen(endsWith("__init__.py")),
            map_(toModulePath),
            forEach(importModule),
        ],
    )


def importModule(modulePath):
    try:
        spec = importlib.util.find_spec(modulePath)
        testModule = importlib.util.module_from_spec(spec)
        testModule.test = test
        testModule.suite = suite
        spec.loader.exec_module(testModule)
    except Exception as e:
        raise Exception(f"Error occurred while importing '{modulePath}'") from e


def makeCmd(someStr, beforeOrAfter):
    fname = someStr.strip()
    _base, ext = path.splitext(fname)

    # not sure why this is necessary :(
    relativeFname = path.join(".", fname)

    if ext == ".py":
        return lambda: run([sys.executable, relativeFname])
    elif ext == ".sh":
        return lambda: run([relativeFname])

    raise_(
        ValueError,
        f"""
        incorrect value found in pyproject.toml
          -> tool.simple_test
          -> {beforeOrAfter}

        file extension must be .py or .sh.
        extension found: '{ext}'

        full path to pyproject.toml:
        {path.join(os.getcwd(), "pyproject.toml")}
        """,
    )


def recursiveGlob(globStr):
    return glob(globStr, recursive=True)


def toModulePath(filePathFromTestsDir):
    return passThrough(
        filePathFromTestsDir, [removeExtension, split(os.sep), joinWith(".")]
    )


def removeExtension(filePath):
    return os.path.splitext(filePath)[0]



import re
import sys
import unicodedata

xrange = range


def _charIsSep(aChar):
    return (
        not _charIsUpper(aChar)
        and not _charIsLower(aChar)
        and not _charIsNumberDecimalDigit(aChar)
    )


def _isSep(aString):
    return len(aString) == 1 and _charIsSep(aString)


def _charIsNumberDecimalDigit(aChar):
    return unicodedata.category(aChar) == "Nd"


def _charIsLower(aChar):
    return unicodedata.category(aChar) == "Ll"


def _charIsUpper(aChar):
    return unicodedata.category(aChar) == "Lu"


def _isUpper(aString):
    return len(aString) == 1 and _charIsUpper(aString)


def _isValidAcronym(aString):
    if len(aString) == 0:
        return False

    for aChar in aString:
        if _charIsSep(aChar):
            return False

    return True


def _determine_case(was_upper, words, string):
    """
    Determine case type of string.

    Arguments:
        was_upper {[type]} -- [description]
        words {[type]} -- [description]
        string {[type]} -- [description]

    Returns:
        - upper: All words are upper-case.
        - lower: All words are lower-case.
        - pascal: All words are title-case or upper-case. Note that the
                  stringiable may still have separators.
        - camel: First word is lower-case, the rest are title-case or
                 upper-case. stringiable may still have separators.
        - mixed: Any other mixing of word casing. Never occurs if there are
                 no separators.
        - unknown: stringiable contains no words.

    """
    case_type = "unknown"
    if was_upper:
        case_type = "upper"
    elif string.islower():
        case_type = "lower"
    elif len(words) > 0:
        camel_case = words[0].islower()
        pascal_case = words[0].istitle() or words[0].isupper()

        if camel_case or pascal_case:
            for word in words[1:]:
                c = word.istitle() or word.isupper()
                camel_case &= c
                pascal_case &= c
                if not c:
                    break

        if camel_case:
            case_type = "camel"
        elif pascal_case:
            case_type = "pascal"
        else:
            case_type = "mixed"

    return case_type


def _advanced_acronym_detection(s, i, words, acronyms):
    """
    Detect acronyms by checking against a list of acronyms.

    Check a run of words represented by the range [s, i].
    Return last index of new word groups.
    """
    # Combine each letter into single string.
    acstr = "".join(words[s:i])

    # List of ranges representing found acronyms.
    range_list = []
    # Set of remaining letters.
    not_range = set(range(len(acstr)))

    # Search for each acronym in acstr.
    for acronym in acronyms:
        # TODO: Sanitize acronyms to include only letters.
        rac = re.compile(acronym)

        # Loop until all instances of the acronym are found,
        # instead of just the first.
        n = 0
        while True:
            m = rac.search(acstr, n)
            if not m:
                break

            a, b = m.start(), m.end()
            n = b

            # Make sure found acronym doesn't overlap with others.
            ok = True
            for r in range_list:
                if a < r[1] and b > r[0]:
                    ok = False
                    break

            if ok:
                range_list.append((a, b))
                for j in xrange(a, b):
                    not_range.remove(j)

    # Add remaining letters as ranges.
    for nr in not_range:
        range_list.append((nr, nr + 1))

    # No ranges will overlap, so it's safe to sort by lower bound,
    # which sort() will do by default.
    range_list.sort()

    # Remove original letters in word list.
    for _ in xrange(s, i):
        del words[s]

    # Replace them with new word grouping.
    for j in xrange(len(range_list)):
        r = range_list[j]
        words.insert(s + j, acstr[r[0] : r[1]])

    return s + len(range_list) - 1


def _simple_acronym_detection(s, i, words, *args):
    """Detect acronyms based on runs of upper-case letters."""
    # Combine each letter into a single string.
    acronym = "".join(words[s:i])

    # Remove original letters in word list.
    for _ in xrange(s, i):
        del words[s]

    # Replace them with new word grouping.
    words.insert(s, "".join(acronym))

    return s


class InvalidAcronymError(Exception):
    """Raise when acronym fails validation."""

    def __init__(self, acronym):
        super(InvalidAcronymError, self).__init__()
        m = "Case Conversion: acronym '{}' is invalid."
        print(m.format(acronym))


def _sanitize_acronyms(unsafe_acronyms):
    """
    Check acronyms against regex.

    Normalize valid acronyms to upper-case.
    If an invalid acronym is encountered raise InvalidAcronymError.
    """
    acronyms = []
    for a in unsafe_acronyms:
        if _isValidAcronym(a):
            acronyms.append(a.upper())
        else:
            raise InvalidAcronymError(a)
    return acronyms


def _normalize_words(words, acronyms):
    """Normalize case of each word to PascalCase."""
    for i, _ in enumerate(words):
        # if detect_acronyms:
        if words[i].upper() in acronyms:
            # Convert known acronyms to upper-case.
            words[i] = words[i].upper()
        else:
            # Fallback behavior: Preserve case on upper-case words.
            if not words[i].isupper():
                words[i] = words[i].capitalize()
    return words


def _separate_words(string):
    """
    Segment string on separator into list of words.

    Arguments:
        string -- the string we want to process

    Returns:
        words -- list of words the string got minced to
        separator -- the separator char intersecting words
        was_upper -- whether string happened to be upper-case
    """
    words = []
    separator = ""

    # Index of current character. Initially 1 because we don't want to check
    # if the 0th character is a boundary.
    i = 1
    # Index of first character in a sequence
    s = 0
    # Previous character.
    p = string[0:1]

    # Treat an all-caps stringiable as lower-case, so that every letter isn't
    # counted as a boundary.
    was_upper = False
    if string.isupper():
        string = string.lower()
        was_upper = True

    # Iterate over each character, checking for boundaries, or places where
    # the stringiable should divided.
    while i <= len(string):
        c = string[i : i + 1]

        split = False
        if i < len(string):
            # Detect upper-case letter as boundary.
            if _charIsUpper(c):
                split = True
            # Detect transition from separator to not separator.
            elif not _charIsSep(c) and _charIsSep(p):
                split = True
            # Detect transition not separator to separator.
            elif _charIsSep(c) and not _charIsSep(p):
                split = True
        else:
            # The loop goes one extra iteration so that it can handle the
            # remaining text after the last boundary.
            split = True

        if split:
            if not _charIsSep(p):
                words.append(string[s:i])
            else:
                # stringiable contains at least one separator.
                # Use the first one as the stringiable's primary separator.
                if not separator:
                    separator = string[s : s + 1]

                # Use None to indicate a separator in the word list.
                words.append(None)
                # If separators weren't included in the list, then breaks
                # between upper-case sequences ("AAA_BBB") would be
                # disregarded; the letter-run detector would count them as one
                # sequence ("AAABBB").
            s = i

        i += 1
        p = c

    return words, separator, was_upper


def parse_case(string, acronyms=None, preserve_case=False):
    """
    Parse a stringiable into a list of words.

    Also returns the case type, which can be one of the following:
        - upper: All words are upper-case.
        - lower: All words are lower-case.
        - pascal: All words are title-case or upper-case. Note that the
                  stringiable may still have separators.
        - camel: First word is lower-case, the rest are title-case or
                 upper-case. stringiable may still have separators.
        - mixed: Any other mixing of word casing. Never occurs if there are
                 no separators.
        - unknown: stringiable contains no words.

    Also returns the first separator character, or False if there isn't one.

    """
    words, separator, was_upper = _separate_words(string)

    if acronyms:
        # Use advanced acronym detection with list
        acronyms = _sanitize_acronyms(acronyms)
        check_acronym = _advanced_acronym_detection
    else:
        acronyms = []
        # Fallback to simple acronym detection.
        check_acronym = _simple_acronym_detection

    # Letter-run detector

    # Index of current word.
    i = 0
    # Index of first letter in run.
    s = None

    # Find runs of single upper-case letters.
    while i < len(words):
        word = words[i]
        if word is not None and _isUpper(word):
            if s is None:
                s = i
        elif s is not None:
            i = check_acronym(s, i, words, acronyms) + 1
            s = None

        i += 1

    if s is not None:
        check_acronym(s, i, words, acronyms)

    # Separators are no longer needed, so they can be removed. They *should*
    # be removed, since it's supposed to be a *word* list.
    words = [w for w in words if w is not None]

    # Determine case type.
    case_type = _determine_case(was_upper, words, string)

    if preserve_case:
        if was_upper:
            words = [w.upper() for w in words]
    else:
        words = _normalize_words(words, acronyms)

    return words, case_type, separator



# ------- #
# Imports #
# ------- #

from .internal.makeCallFn import makeCallFn
from .internal.getTypedResult import getTypedResult


# ---- #
# Init #
# ---- #

moduleName = "keepWhen"


# ---- #
# Main #
# ---- #


def keepWhen(predicate):
    shouldKeep = makeCallFn(predicate, 3, moduleName)

    def keepWhen_inner(collection):
        typedKeepWhen = getTypedResult(collection, _typeToKeepWhen, moduleName)
        return typedKeepWhen(shouldKeep, collection)

    return keepWhen_inner


# ------- #
# Helpers #
# ------- #


def keepWhen_list(shouldKeep, aList):
    result = []
    for idx, el in enumerate(aList):
        if shouldKeep(el, idx, aList):
            result.append(el)

    return result


def keepWhen_dict(shouldKeep, aDict):
    result = {}
    for key, val in aDict.items():
        if shouldKeep(val, key, aDict):
            result[key] = val

    return result


_typeToKeepWhen = {list: keepWhen_list, dict: keepWhen_dict}



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace as o
from ...fns import assign, map_, mAssign, mSet


# ---- #
# Main #
# ---- #


def applyIndentLevels(state):
    rootTests = map_(addIndentLevelToTest(0))(state.tests)
    rootSuites = map_(addIndentLevelToSuite(0))(state.suites)
    return assign(o(tests=rootTests, suites=rootSuites))(state)


# ------- #
# Helpers #
# ------- #


def addIndentLevelToTest(lvl):
    def applyIndentLevel_inner(aTest):
        return mSet("indentLevel", lvl)(aTest)

    return applyIndentLevel_inner


def addIndentLevelToSuite(lvl):
    def addIndentLevelToSuite_inner(aSuite):
        indentLevel = lvl
        tests = map_(addIndentLevelToTest(lvl + 1))(aSuite.tests)
        suites = map_(addIndentLevelToSuite(lvl + 1))(aSuite.suites)
        return mAssign(o(indentLevel=indentLevel, tests=tests, suites=suites))(
            aSuite
        )

    return addIndentLevelToSuite_inner



from textwrap import dedent


def throw(errorClass, message, *, fromException=None):
    if fromException is None:
        raise errorClass(dedent(message))
    else:
        raise errorClass(dedent(message)) from fromException



# ------- #
# Imports #
# ------- #

from ..fns import forEach
from .utils import o, x


# ---- #
# Main #
# ---- #


def reportTest(aTest, idx, tests):
    indent = " " * 4 * aTest.indentLevel
    if aTest.succeeded:
        print(f"{indent}{aTest.label} {o}")
    else:
        print(f"{indent}{aTest.label} {x}")

    if (
        aTest.parentSuite
        and isLast(idx, tests)
        and not isFinalSuite(aTest.parentSuite)
    ):
        propagateBlankLineAfter(aTest.parentSuite)
        print()


def reportSuite(aSuite, idx, suites):
    if (
        aSuite.parentSuite is None
        and idx != 0
        and not suites[idx - 1].hasBlankLineAfter
    ):
        print()

    indent = " " * 4 * aSuite.indentLevel
    if aSuite.succeeded:
        print(f"{indent}{aSuite.label} {o}")
        # we only want to report failing tests, and if a suite succeeded then
        #   it doesn't contain any failing tests
        forEach(reportSuite)(aSuite.suites)
    else:
        print(f"{indent}{aSuite.label} {x}")
        forEach(reportTest)(aSuite.tests)
        forEach(reportSuite)(aSuite.suites)


# ------- #
# Helpers #
# ------- #


#
# TODO: initialize this state separate from the reporting logic similar to
#   applyIndentLevels.  That way the reporter can just check the properties on
#   the state instead of modifying it mid-run.
#
# this method is necessary due to two formatting preferences
#   1. root suites should be separated by one blank line
#   2. there should be a blank line between a test and the following suite
#
# these two preferences conflict when a blank line follows a failed test where
#   the following suite is a root suite.  By propagating the "hasBlankLineAfter"
#   state we can check that on root suites to prevent two blank lines.
#
def propagateBlankLineAfter(aSuite):
    if not aSuite.parentSuite:
        aSuite.hasBlankLineAfter = True
        return

    if isLastSiblingSuite(aSuite):
        propagateBlankLineAfter(aSuite.parentSuite)


def isLastSiblingSuite(suite):
    parentSuite = suite.parentSuite
    if parentSuite:
        siblingSuites = parentSuite.suites
    else:
        siblingSuites = suite.parentState.suites

    indexOfSuite = siblingSuites.index(suite)
    return isLast(indexOfSuite, siblingSuites)


def isLast(idx, collection):
    return idx == len(collection) - 1


def isFinalSuite(aSuite):
    if not aSuite.parentSuite:
        return isLastSiblingSuite(aSuite)
    elif isLastSiblingSuite(aSuite):
        return isFinalSuite(aSuite.parentSuite)



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult
from .mAppendAll import typeToMAppendAll


# ---- #
# Main #
# ---- #


def mAppendAllTo(collection):
    def mAppendAllTo_inner(collectionToAppend):
        typedMAppendAllTo = getTypedResult(
            collection, typeToMAppendAll, mAppendAllTo.__name__
        )
        return typedMAppendAllTo(collectionToAppend, collection)

    return mAppendAllTo_inner



from .internal.getNumPositionalParams import getNumPositionalParams


def not_(predicate):
    if not callable(predicate):
        raise ValueError("'not' requires its argument to be callable")

    numPositionalParams = getNumPositionalParams(predicate, "not_")

    def not_inner(*args, **kwargs):
        return not predicate(*args[:numPositionalParams], **kwargs)

    return not_inner



from ..throw import throw


def validateStr(maybeString, fnName):
    typePassed = type(maybeString)
    if typePassed is not str:
        throw(
            ValueError,
            f"""\
            {fnName} requires a string
            type passed: {typePassed}
            """,
        )



from .. import chalk
from .utils import makeTestResults


def runTests():
    expectedErrorResults = makeTestResults("expected errors")
    e = expectedErrorResults.errors

    try:
        chalk("error")
        e.append('chalk("error")')
    except:
        pass

    return [expectedErrorResults]



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def flatten(collection):
    typedFlatten = getTypedResult(collection, _typeToFlatten, "flatten")
    return typedFlatten(collection)


# ------- #
# Helpers #
# ------- #


def flatten_list(aList):
    result = []
    for maybeInnerList in aList:
        if isinstance(maybeInnerList, list):
            for el in maybeInnerList:
                result.append(el)
        else:
            result.append(maybeInnerList)

    return result


_typeToFlatten = {list: flatten_list}



# ------- #
# Imports #
# ------- #

from math import inf

from .getNumPositionalParams import getNumPositionalParams
from ..throw import throw


# ---- #
# Main #
# ---- #


def makeCallFn(fn, maxParams, callerName):
    numPositionalParams = getNumPositionalParams(fn, callerName)

    if numPositionalParams is inf:
        numPositionalParams = maxParams

    if numPositionalParams > maxParams:
        throw(
            ValueError,
            f"""
            {callerName} can only take functions with up to {maxParams} params
            the function passed '{fn.__name__}' takes {numPositionalParams}
            """,
        )

    return lambda *args: fn(*args[:numPositionalParams])



# ------- #
# Imports #
# ------- #

from .internal.makeCallFn import makeCallFn
from .internal.getTypedResult import getTypedResult


# ---- #
# Init #
# ---- #

moduleName = "forEach"


# ---- #
# Main #
# ---- #


def forEach(fn):
    callFn = makeCallFn(fn, 3, moduleName)

    def forEach_inner(collection):
        typedForEach = getTypedResult(collection, _typeToForEach, moduleName)
        return typedForEach(callFn, collection)

    return forEach_inner


# ------- #
# Helpers #
# ------- #


def forEach_list(callFn, aList):
    for idx, el in enumerate(aList):
        callFn(el, idx, aList)

    return aList


_typeToForEach = {list: forEach_list}



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def appendOne(el):
    def appendOne_inner(collection):
        typedAppendOne = getTypedResult(
            collection, typeToAppendOne, appendOne.__name__
        )
        return typedAppendOne(el, collection)

    return appendOne_inner


# ------- #
# Helpers #
# ------- #


def appendOne_list(el, aList):
    return aList + [el]


typeToAppendOne = {list: appendOne_list}



# ------- #
# Imports #
# ------- #

from ..src.data import first8Colors, second8Colors
from .. import *  # noqa: F403
from ..src import utils as _
from .utils import makeTestResults


# ---- #
# Main #
# ---- #


def runTests():
    aliasResults = makeTestResults("alias colors")
    foregroundResults = makeTestResults("foreground colors")
    backgroundResults = makeTestResults("background colors")

    aliasResults.errors = testAliases()

    foregroundResults.errors = testColors(first8Colors, 30) + testColors(
        second8Colors, 90
    )

    backgroundResults.errors = testColors(
        first8Colors, 40, isBackground=True
    ) + testColors(second8Colors, 100, isBackground=True)

    return [foregroundResults, backgroundResults, aliasResults]


# ------- #
# Helpers #
# ------- #


def testAliases():
    return (
        testAColor([], "gray", 90, False)
        + testAColor([], "gray", 100, True)
        + testAColor([], "grey", 90, False)
        + testAColor([], "grey", 100, True)
    )


def testAColor(errors, name, code, isBackground):
    if isBackground:
        name = f"bg{_.upperFirst(name)}"

    expected = f"\x1b[{code}msuccess\x1b[0m"
    if globals()[name]("success") != expected:
        errors.append(f'{name}("success")')
    if getattr(chalk, name)("success") != expected:  # noqa: F405
        errors.append(f'chalk.{name}("success")')

    return errors


def testColors(colors, startCode, *, isBackground=False):
    def testAColorWrapper(errors, name, idx):
        return testAColor(errors, name, idx + startCode, isBackground)

    return _.reduce(testAColorWrapper, [])(colors)



# ------- #
# Imports #
# ------- #

from ...fns import passThrough
from .applyIndentLevels import applyIndentLevels
from .removeSucceededTestsInSuites import removeSucceededTestsInSuites


# ---- #
# Main #
# ---- #


def initRootState(state):
    return passThrough(
        state,
        [
            removeSucceededTestsInSuites,
            applyIndentLevels,
            initializeHasBlankLineAfter,
            initializeParentState,
        ],
    )


# ------- #
# Helpers #
# ------- #


def initializeHasBlankLineAfter(state):
    for suite in state.suites:
        suite.hasBlankLineAfter = False

    return state


def initializeParentState(state):
    for suite in state.suites:
        suite.parentState = state

    for test in state.tests:
        test.parentState = state

    return state



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult
from types import SimpleNamespace
from copy import copy


# ---- #
# Main #
# ---- #


def assign(collection):
    fnName = assign.__name__
    assignFn = getTypedResult(collection, typeToAssign, fnName)

    return assignFn(collection)


# ------- #
# Helpers #
# ------- #


def assign_simpleNamespace(primary):
    def assign_simpleNamespace_inner(secondary):
        result = copy(primary)

        for k, v in secondary.__dict__.items():
            if k not in result.__dict__:
                setattr(result, k, v)

        return result

    return assign_simpleNamespace_inner


def assign_dict(primary):
    def assign_dict_inner(secondary):
        result = copy(primary)

        for k, v in secondary.items():
            if k not in result:
                result[k] = v

        return result

    return assign_dict_inner


typeToAssign = {dict: assign_dict, SimpleNamespace: assign_simpleNamespace}



# ------- #
# Imports #
# ------- #

from traceback import format_exc
from .fns import forEach, invoke


# ---- #
# Main #
# ---- #


def runAllTests(stateOrSuite):
    stateOrSuite.before()
    forEach(runTest)(stateOrSuite.tests)
    forEach(runAllTests)(stateOrSuite.suites)
    stateOrSuite.after()


# ------- #
# Helpers #
# ------- #


def runTest(aTest):
    parent = aTest.parentSuite or aTest.rootState
    try:
        forEach(invoke)(parent.beforeEach)
        aTest.before()
        aTest.fn()
        aTest.succeeded = True
    except Exception as e:
        aTest.succeeded = False
        aTest.rootState.succeeded = False
        propagateFailure(aTest.parentSuite)

        # I can't think of a good condition for when to attach the stacktrace to
        #   the reporter output.  If we make it opt-in from the expect library
        #   then that forces every error to have metadata that's only meaningful
        #   for simple-test.  Alternatively making it opt-in from simple-test
        #   means every time a library decides they don't want stack trace info
        #   attached we have to update this condition.  So I'm punting for now
        #   in hopes an actual use-case will clarify a solution.
        if type(e).__name__ != "ExpectFailedError":
            aTest.formattedException = format_exc()

        aTest.error = e
    finally:
        aTest.after()
        forEach(invoke)(parent.afterEach)


def propagateFailure(aSuite):
    if aSuite is None:
        return

    if aSuite.succeeded:
        aSuite.succeeded = False
        propagateFailure(aSuite.parentSuite)



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def mAppendAll(collectionToAppend):
    def mAppendAll_inner(collection):
        typedMAppendAll = getTypedResult(
            collection, typeToMAppendAll, mAppendAll.__name__
        )
        return typedMAppendAll(collectionToAppend, collection)

    return mAppendAll_inner


# ------- #
# Helpers #
# ------- #


def mAppendAll_list(listToAppend, aList):
    aList += listToAppend
    return aList


typeToMAppendAll = {list: mAppendAll_list}



from .parseArgs import parseArgs
from .runProcess import runProcess
import sys


def printErr(msg):
    print(msg, file=sys.stderr)


kwargs = parseArgs(*sys.argv[1:]).__dict__
result = runProcess(**kwargs)

if result.stdout:
    print(result.stdout)

if result.stderr:
    printErr(result.stderr)

exit(result.code)



from __future__ import absolute_import

from .case_conversion import (
    camelcase, pascalcase, snakecase, dashcase, kebabcase, spinalcase,
    constcase, screaming_snakecase, dotcase, separate_words, slashcase,
    backslashcase)



from copy import copy, deepcopy
from types import SimpleNamespace as o
from .fns import appendOne, assign, noop

_initialState = o(
    tests=[],
    suites=[],
    currentSuite=None,
    testsFound=False,
    succeeded=True,
    after=noop,
    afterEach=[],
    before=noop,
    beforeEach=[],
)


state = deepcopy(_initialState)


def _getState():
    return state


def initState():
    global state
    state = deepcopy(_initialState)
    return state


def _getCommon(*, label, fn, parentSuite, rootState, after, before):
    return o(
        label=label,
        fn=fn,
        parentSuite=parentSuite,
        rootState=rootState,
        after=after,
        before=before,
    )


def addSuite(label, after, afterEach, before, beforeEach, fn):
    currentSuite = state.currentSuite
    parent = currentSuite or state

    if afterEach is noop:
        afterEach = copy(parent.afterEach)
    else:
        afterEach = appendOne(afterEach)(parent.afterEach)

    if beforeEach is noop:
        beforeEach = copy(parent.beforeEach)
    else:
        beforeEach = appendOne(beforeEach)(parent.beforeEach)

    newSuite = assign(
        o(
            tests=[],
            suites=[],
            succeeded=True,
            afterEach=afterEach,
            beforeEach=beforeEach,
        )
    )(
        _getCommon(
            label=label,
            fn=fn,
            parentSuite=currentSuite,
            rootState=state,
            after=after,
            before=before,
        )
    )

    parent.suites.append(newSuite)

    return state


def addTest(label, after, before, fn):
    global state
    currentSuite = state.currentSuite
    test = _getCommon(
        label=label,
        fn=fn,
        parentSuite=currentSuite,
        rootState=state,
        after=after,
        before=before,
    )

    if currentSuite is None:
        state.tests.append(test)
    else:
        currentSuite.tests.append(test)

    state.testsFound = True

    return state



from .internal.validateStr import validateStr


def startsWith(prefix):
    validateStr(prefix, "startsWith")

    def startsWith_inner(fullStr):
        validateStr(fullStr, "startsWith_inner")
        return fullStr.startswith(prefix)

    return startsWith_inner



from ..simple_chalk.src import utils as _
from .src.makeAChalkInstance import makeAChalkInstance as newChalk

chalk = newChalk()

__all__ = ["chalk", "newChalk"]


def doesntStartWith(prefix):
    return _.not_(_.startsWith(prefix))


chalkClass = type(chalk)


def setGlobal(attrName):
    globals()[attrName] = getattr(chalk, attrName)
    __all__.append(attrName)


_.passThrough(
    chalk,
    [
        type,
        vars,
        _.invokeAttr("keys"),
        list,
        _.keepWhen(doesntStartWith("_")),
        _.forEach(setGlobal),
    ],
)



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def join(separator):
    def join_inner(collection):
        typedJoin = getTypedResult(collection, _typeToJoin, "join")
        return typedJoin(separator, collection)

    return join_inner


# ------- #
# Helpers #
# ------- #


def join_list(separator, aList):
    return separator.join(aList)


_typeToJoin = {list: join_list}



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace as o
from ...fns import mAssign, discardWhen, map_, set_


# ---- #
# Main #
# ---- #


def removeSucceededTestsInSuites(state):
    rootSuites = map_(removeSucceededTests)(state.suites)
    return set_("suites", rootSuites)(state)


# ------- #
# Helpers #
# ------- #


def isSucceeded(aTest):
    return aTest.succeeded


def removeSucceededTests(aSuite):
    tests = discardWhen(isSucceeded)(aSuite.tests)
    suites = map_(removeSucceededTests)(aSuite.suites)
    return mAssign(o(tests=tests, suites=suites))(aSuite)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['po',
 'po.simple_test',
 'po.simple_test._vendor',
 'po.simple_test._vendor.po',
 'po.simple_test._vendor.po.case_conversion',
 'po.simple_test._vendor.simple_test_process',
 'po.simple_test._vendor.simple_test_process._vendor',
 'po.simple_test._vendor.simple_test_process._vendor.po',
 'po.simple_test._vendor.simple_test_process._vendor.po.case_conversion',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.simple_chalk',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.simple_chalk.src',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.simple_chalk.src.utils',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.simple_chalk.src.utils.internal',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.simple_chalk.tests',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent._vendor',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent._vendor.wrapt',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent.fns',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent.fns.decorators',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent.fns.internal',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.wrapt',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter.fns',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter.fns.decorators',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter.fns.internal',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter.report',
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter.report.initRootState',
 'po.simple_test._vendor.simple_test_process._vendor.tedent',
 'po.simple_test._vendor.simple_test_process._vendor.tedent._vendor',
 'po.simple_test._vendor.simple_test_process._vendor.tedent._vendor.wrapt',
 'po.simple_test._vendor.simple_test_process._vendor.tedent.fns',
 'po.simple_test._vendor.simple_test_process._vendor.tedent.fns.decorators',
 'po.simple_test._vendor.simple_test_process._vendor.tedent.fns.internal',
 'po.simple_test._vendor.simple_test_process._vendor.toml',
 'po.simple_test._vendor.simple_test_process._vendor.wrapt',
 'po.simple_test._vendor.simple_test_process.fns',
 'po.simple_test._vendor.simple_test_process.fns.decorators',
 'po.simple_test._vendor.simple_test_process.fns.internal',
 'po.simple_test._vendor.tedent',
 'po.simple_test._vendor.tedent._vendor',
 'po.simple_test._vendor.tedent._vendor.wrapt',
 'po.simple_test._vendor.tedent.fns',
 'po.simple_test._vendor.tedent.fns.decorators',
 'po.simple_test._vendor.tedent.fns.internal',
 'po.simple_test._vendor.wrapt',
 'po.simple_test.cli',
 'po.simple_test.fns',
 'po.simple_test.fns.decorators',
 'po.simple_test.fns.internal',
 'po.simple_test.run']

package_data = \
{'': ['*'],
 'po.simple_test._vendor': ['ordered_set-3.1.1.dist-info/*',
                            'po.case_conversion-0.3.0.dist-info/*',
                            'simple_test_process-0.5.0.dist-info/*',
                            'tedent-0.1.5.dist-info/*',
                            'wrapt-1.10.11.dist-info/*'],
 'po.simple_test._vendor.simple_test_process._vendor': ['ordered_set-3.1.1.dist-info/*',
                                                        'po.case_conversion-0.2.0.dist-info/*',
                                                        'simple_test_default_reporter-0.2.1.dist-info/*',
                                                        'tedent-0.1.5.dist-info/*',
                                                        'toml-0.10.0.dist-info/*',
                                                        'wrapt-1.10.11.dist-info/*'],
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor': ['ordered_set-3.1.1.dist-info/*',
                                                                                             'simple_chalk-0.1.0.dist-info/*',
                                                                                             'tedent-0.1.5.dist-info/*',
                                                                                             'wrapt-1.10.11.dist-info/*'],
 'po.simple_test._vendor.simple_test_process._vendor.simple_test_default_reporter._vendor.tedent._vendor': ['ordered_set-3.1.1-py3.7.egg-info/*',
                                                                                                            'wrapt-1.11.1.dist-info/*'],
 'po.simple_test._vendor.simple_test_process._vendor.tedent._vendor': ['ordered_set-3.1.1-py3.7.egg-info/*',
                                                                       'wrapt-1.11.1.dist-info/*'],
 'po.simple_test._vendor.tedent._vendor': ['ordered_set-3.1.1-py3.7.egg-info/*',
                                           'wrapt-1.11.1.dist-info/*']}

entry_points = \
{'console_scripts': ['simple-test = po.simple_test.script:main']}

setup_kwargs = {
    'name': 'po.simple-test',
    'version': '0.4.2',
    'description': 'A simple test runner',
    'long_description': 'I will document this module if people will find it helpful.  Currently\nsimple_test is just for my own personal use.\n',
    'author': 'Philip Olson',
    'author_email': 'philip.olson@pm.me',
    'url': 'https://github.com/olsonpm/py_simple-test',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys

PYTHON = sys.version_info[0]

if 3 == PYTHON:
    # Python 3 and ST3
    from . import case_parse
else:
    # Python 2 and ST2
    import case_parse


def camelcase(text, acronyms=None):
    """Return text in camelCase style.

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> camelcase("hello world")
    'helloWorld'
    >>> camelcase("HELLO_HTML_WORLD", True, ["HTML"])
    'helloHTMLWorld'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms)
    if words:
        words[0] = words[0].lower()
    return ''.join(words)


def pascalcase(text, acronyms=None):
    """Return text in PascalCase style (aka MixedCase).

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> pascalcase("hello world")
    'HelloWorld'
    >>> pascalcase("HELLO_HTML_WORLD", True, ["HTML"])
    'HelloHTMLWorld'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms)
    return ''.join(words)


def snakecase(text, acronyms=None):
    """Return text in snake_case style.

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> snakecase("hello world")
    'hello_world'
    >>> snakecase("HelloHTMLWorld", True, ["HTML"])
    'hello_html_world'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms)
    return '_'.join([w.lower() for w in words])


def dashcase(text, acronyms=None):
    """Return text in dash-case style (aka kebab-case, spinal-case).

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> dashcase("hello world")
    'hello-world'
    >>> dashcase("HelloHTMLWorld", True, ["HTML"])
    'hello-html-world'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms)
    return '-'.join([w.lower() for w in words])


def kebabcase(text, acronyms=None):
    """Return text in kebab-case style (aka snake-case, spinal-case).

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> kebabcase("hello world")
    'hello-world'
    >>> kebabcase("HelloHTMLWorld", True, ["HTML"])
    'hello-html-world'
    """
    return dashcase(text, acronyms)


def spinalcase(text, acronyms=None):
    """Return text in spinal-case style (aka snake-case, kebab-case).

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> spinalcase("hello world")
    'hello-world'
    >>> spinalcase("HELLO_HTML_WORLD", True, ["HTML"])
    'hello-html-world'
    """
    return dashcase(text, acronyms)


def constcase(text, acronyms=None):
    """Return text in CONST_CASE style (aka SCREAMING_SNAKE_CASE).

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> constcase("hello world")
    'HELLO_WORLD'
    >>> constcase("helloHTMLWorld", True, ["HTML"])
    'HELLO_HTML_WORLD'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms)
    return '_'.join([w.upper() for w in words])


def screaming_snakecase(text, acronyms=None):
    """Return text in SCREAMING_SNAKE_CASE style (aka CONST_CASE).

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> screaming_snakecase("hello world")
    'HELLO_WORLD'
    >>> screaming_snakecase("helloHTMLWorld", True, ["HTML"])
    'HELLO_HTML_WORLD'
    """
    return constcase(text, acronyms)


def dotcase(text, acronyms=None):
    """Return text in dot.case style.

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> dotcase("hello world")
    'hello.world'
    >>> dotcase("helloHTMLWorld", True, ["HTML"])
    'hello.html.world'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms)
    return '.'.join([w.lower() for w in words])


def separate_words(text, acronyms=None):
    """Return text in "seperate words" style.

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> separate_words("HELLO_WORLD")
    'HELLO WORLD'
    >>> separate_words("helloHTMLWorld", True, ["HTML"])
    'hello HTML World'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms, preserve_case=True)
    return ' '.join(words)


def slashcase(text, acronyms=None):
    """Return text in slash/case style.

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> slashcase("HELLO_WORLD")
    'HELLO/WORLD'
    >>> slashcase("helloHTMLWorld", True, ["HTML"])
    'hello/HTML/World'
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms, preserve_case=True)
    return '/'.join(words)


def backslashcase(text, acronyms=None):
    """Return text in backslash\case style.

    Args:
        text: input string to convert case
        detect_acronyms: should attempt to detect acronyms
        acronyms: a list of acronyms to detect

    >>> backslashcase("HELLO_WORLD") == r'HELLO\WORLD'
    True
    >>> backslashcase("helloHTMLWorld", True, ["HTML"]) == r'hello\HTML\World'
    True
    """
    words, _case, _sep = case_parse.parse_case(text, acronyms, preserve_case=True)
    return '\\'.join(words)



# ------- #
# Imports #
# ------- #

from os import path
from ._vendor.tedent import tedent
from traceback import format_exc
from types import SimpleNamespace as o
from .onlyKeepGreppedTests import onlyKeepGreppedTests
from .runAllTests import runAllTests
from .state import initState
from .utils import gatherTests, importTests, makeCmd, twoLineSeps
from .validateAndGetReportFn import validateAndGetReportFn
import os
from ._vendor import toml
from ._vendor import simple_test_default_reporter

from .fns import forEach, iif, isEmpty, joinWith, map_, noop, passThrough, prependStr


# ---- #
# Main #
# ---- #


def runProcess(*, reporter, silent, grepArgs):
    try:
        silent = silent == "True"
        cliResult = o(stderr=None, stdout=None, code=None)
        after = noop
        before = noop

        if not silent:
            if reporter == "None":
                report = simple_test_default_reporter.report
            else:
                validationResult = validateAndGetReportFn(reporter, silent, cliResult)

                if validationResult.hasError:
                    return validationResult.cliResult

                report = validationResult.report

            reportOpts = None
            if path.isfile("pyproject.toml"):
                allProjectSettings = toml.load("pyproject.toml")
                result = getValueAtPath(allProjectSettings, ["tool", reporter])
                if result.hasValue:
                    reportOpts = result.value

                result = getValueAtPath(allProjectSettings, ["tool", "simple_test"])
                if result.hasValue:
                    if "before" in result.value:
                        before = makeCmd(result.value["before"], "before")

                    if "after" in result.value:
                        after = makeCmd(result.value["after"], "after")

        state = initState()
        state.before = before
        state.after = after
        importTests()

        forEach(gatherTests)(state.suites)
        onlyKeepGreppedTests(state, grepArgs)
        runAllTests(state)

        if not state.testsFound:
            if not silent:
                cliResult.stderr = tedent(
                    f"""
                    No tests were found in any python files under the project's
                    tests directory: '{path.join(os.getcwd(), 'tests')}'

                    Remember you define tests by decorating a function with
                    @test("test label")
                    """
                )

            cliResult.code = 2
            return cliResult

        if isEmpty(state.tests) and isEmpty(state.suites):
            # grepArgs must contain something if this code is reached
            if not silent:
                cliResult.stderr = (
                    "Your grep options failed to match any suites or tests"
                )

            cliResult.code = 2
            return cliResult

        if not silent:
            if reportOpts is None:
                report(state)
            else:
                report(state, reportOpts)

        cliResult.code = iif(state.succeeded, 0, 1)
        return cliResult

    except Exception:
        if not silent:
            cliResult.stderr = (
                os.linesep
                + "An error occurred during simple_test_process"
                + twoLineSeps
                + format_exc()
            )

        cliResult.code = 2
        return cliResult


# ------- #
# Helpers #
# ------- #


def toFormattedGrepArgs(greppedStrings, grepKey):
    return (
        grepKey
        + ":"
        + os.linesep
        + passThrough(greppedStrings, [map_(prependStr("  ")), joinWith(os.linesep)])
    )


def getValueAtPath(aDict, pathToValue):
    result = o(hasValue=None, value=None)
    val = aDict
    for segment in pathToValue:
        if segment not in val:
            result.hasValue = False
            return result

        val = val[segment]

    result.hasValue = True
    result.value = val
    return result



from .state import addSuite
from .fns import noop


def suite(label, *, after=noop, afterEach=noop, before=noop, beforeEach=noop):
    def wrapper(fn):
        addSuite(label, after, afterEach, before, beforeEach, fn)

    return wrapper



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult
from types import SimpleNamespace


# ---- #
# Main #
# ---- #


def mAssign(collection):
    fnName = mAssign.__name__
    assignFn = getTypedResult(collection, typeToAssign, fnName)

    return assignFn(collection)


# ------- #
# Helpers #
# ------- #


def mAssign_simpleNamespace(primary):
    def mAssign_simpleNamespace_inner(secondary):
        for k, v in primary.__dict__.items():
            setattr(secondary, k, v)

        return secondary

    return mAssign_simpleNamespace_inner


def mAssign_dict(primary):
    def mAssign_dict_inner(secondary):
        for k, v in primary.items():
            secondary[k] = v

        return secondary

    return mAssign_dict_inner


typeToAssign = {dict: mAssign_dict, SimpleNamespace: mAssign_simpleNamespace}



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace
from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def getListOfCollectionKeys(collection):
    fnName = getListOfCollectionKeys.__name__
    typedKeys = getTypedResult(collection, typeToKeys, fnName)
    return typedKeys(collection)


# ------- #
# Helpers #
# ------- #


def getListOfCollectionKeys_dict(aDict):
    return list(aDict.keys())


def getListOfCollectionKeys_simpleNamespace(aSimpleNamespace):
    return list(aSimpleNamespace.__dict__.keys())


typeToKeys = {
    dict: getListOfCollectionKeys_dict,
    SimpleNamespace: getListOfCollectionKeys_simpleNamespace,
}
